use std::{
    fs::File,
    io::{BufReader, Read},
    mem::{size_of, transmute},
};

use ::paste::paste;

pub struct OsRandom {
    devurandom: BufReader<File>,
}

impl Default for OsRandom {
    fn default() -> Self {
        Self::new()
    }
}

macro_rules! os_random_get_integer_impls {
    ($($t:ty),*) => {
        $(
            ::paste::paste! {
                pub fn [<get_ $t>](&mut self) -> std::io::Result<$t> {
                    let mut buf = [0u8; size_of::<$t>()];
                    self.devurandom.read_exact(&mut buf)?;
                    Ok(unsafe {transmute::< [u8; size_of::<$t>() ], $t>(buf) })
                }
            }
        )*
    };
}

impl OsRandom {
    pub fn new() -> Self {
        Self {
            devurandom: BufReader::new(File::open("/dev/urandom").unwrap()),
        }
    }

    os_random_get_integer_impls!(u8, u16, u32, u64, u128, usize, i8, i16, i32, i64, i128, isize);
}

impl Read for OsRandom {
    fn read(&mut self, buf: &mut [u8]) -> std::io::Result<usize> {
        self.devurandom.read(buf)
    }
}

macro_rules! os_rng_test_impls {
    ($($t:ty),*) => {
        let mut rng = OsRandom::new();
        $(
            paste! {
                println!("{}", rng.[<get_ $t>]().unwrap());
            }
        )*
    }
}

fn main() {
    os_rng_test_impls!(u8, u16, u32, u64, u128, usize, i8, i16, i32, i64, i128, isize);
}
